// DateFile.cpp : Defines the entry point for the console application.
//

#include "DateFile.h"
#include <windows.h>
#include <direct.h>
//#include <iostream>   // for cin, cout
#include <conio.h>   // for _getch()


#define  VFH(a)   ( a && ( a != INVALID_HANDLE_VALUE ) )
#define  EndBuf(a)   ( a + strlen(a) )

using namespace std;

typedef struct tagFILES {
   string            path; // path of file
   WIN32_FIND_DATA   fd;   // WIN32 information
} FILES, * PFILES;

typedef struct tagFULLPATH {
   TCHAR path[_MAX_PATH];
   TCHAR drive[_MAX_DRIVE];
   TCHAR dir[_MAX_DIR];
   TCHAR fname[_MAX_FNAME];
   TCHAR ext[_MAX_EXT];
} FULLPATH, * PFULLPATH;

typedef vector<string> string_list;
typedef vector<PFILES> file_list;
typedef string_list::iterator siter;
typedef file_list::iterator fiter;

#define  CHKMEM(a)   if( !a ) { printf( "CRITICAL ERROR: MEMORY FAILED!\n" ); pgm_exit(2); }

int bDoUpdate = 0;
string_list files;
file_list   sfiles;
FULLPATH fullpath;

char * Reference = 0;
SYSTEMTIME  Target = { 0 };
FILETIME FileTime = { 0 };
int   bGotTime = 0;
int   bGotDate = 0;
int   bGotRef = 0;
int   bUseSystem = 0;   // use system time, not local time

char ErrMessage[1024];

void  pgm_exit( int ret )
{
   for( fiter fit = sfiles.begin(); fit != sfiles.end(); fit++ ) {
      PFILES pf = *fit;
      delete pf;
   }
   files.clear();
   sfiles.clear();
   exit(ret);
}

void  print_help( void )
{
   printf( "Informatique Rouge - SET FILE DATE/TIME UTILITY - " PGM_DATE "\n" );
   printf( "Useage: DateFile FileSpec [Switches]\n" );
   printf( "Switches: Each preceeded by '-' or '/', and space delimited\n" );
   printf( " -?           This brief help, and exit.\n" );
   printf( " -ddd:mm:yyyy Date to use for update.\n" );
   printf( " -rRefFile    Reference file, used for update date and time.\n" );
   printf( " -s           Use system time for update date and time.\n" );
   printf( " -thh:mm:ss   Time to use for update. Use only 24 hour time.\n" );
   printf( " -u           Update chosen files (Def=Off)\n" );
}

void  give_help( void )
{
   print_help();
   pgm_exit(1);
}

void  no_input( void )
{
   printf( "ERROR: Enter a file, or file mask ...\n" );
   give_help();
}

void  err_cmd( char * cmd )
{
   printf( "ERROR: %s is NOT a valid command!\n", cmd );
   give_help();
}

char * strdupe( char * file )
{
   size_t len = strlen(file) + 1;
   char * cp = new char[len];
   CHKMEM(cp);
   strcpy(cp, file);
   return cp;
}

DWORD GetLastErrorMsg( PTSTR lpm, DWORD dwLen, DWORD dwErr )
{
   PVOID lpMsgBuf = 0;
   DWORD    dwr;

   dwr = FormatMessage( 
      FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		FORMAT_MESSAGE_FROM_SYSTEM | 
		FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL,
		dwErr,   //	GetLastError(),
      MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
      (LPTSTR) &lpMsgBuf,
		0,
		NULL );
   
   //dwr = strlen(lpMsgBuf);
   if( ( dwr == 0 ) || ( dwr >= dwLen ) )
      dwr = (DWORD)-1;
   else
      strcat(lpm, (const char *)lpMsgBuf);

   //	printf("%s:%s\n",lpm,(LPCTSTR)lpMsgBuf);
   // Free the buffer.
   if( lpMsgBuf )
      LocalFree( lpMsgBuf );

   return dwr;
}

PTSTR   GetI64Stg( PULARGE_INTEGER pli )
{
   static TCHAR _s_I64Stg[288];
   TCHAR    buf[64];
   int      i,j,k;
   ULARGE_INTEGER  li;
   PTSTR   lps = _s_I64Stg;
   li = *pli;
   i = sprintf(buf,
      "%I64u",
      li );
   *lps = 0;   // clear any previous
   if( i )  // get its length
   {
      k = 64;
      j = 0;
      lps[k+1] = 0;  // esure ZERO termination
      while( ( i > 0 ) && ( k >= 0 ) )
      {
         i--;     // back up one
         if( j == 3 )   // have we had a set of 3?
         {
            lps[k--] = ',';   // ok, add a comma
            j = 0;            // and restart count of digits
         }
         lps[k--] = buf[i];   // move the buffer digit
         j++;
      }
      k++;  // back to LAST position
      lps = &lps[k]; // pointer to beginning of 'nice" number
   }
   return lps;
}

void  Add_SysTime_Str( PTSTR ps, SYSTEMTIME * pst )
{
   sprintf(EndBuf(ps), "%02d/%02d/%04d %02d:%02d:%02d.%03d",
      pst->wDay, pst->wMonth, pst->wYear, pst->wHour, pst->wMinute,
      pst->wSecond, pst->wMilliseconds );
}

PTSTR  Get_SysTime_Stg( SYSTEMTIME * pst )
{
   static TCHAR _s_syst_buf[32];
   PTSTR ps = _s_syst_buf;
   *ps = 0;
   Add_SysTime_Str( ps, pst );
   return ps;
}

#define  MMINFILNM   32
#define  MMINSIZE    14

PTSTR Get_FD_Stg( WIN32_FIND_DATA * pf )
{
   static char _s_fd_buf[288*2];
   PTSTR cp = _s_fd_buf;
   ULARGE_INTEGER ul;
   SYSTEMTIME  st;

   strcpy(cp, pf->cFileName);
   strcat(cp, " ");  // add at least one space
   while( strlen(cp) < MMINFILNM )
      strcat(cp, " ");

   if( FileTimeToSystemTime( &pf->ftLastWriteTime, &st ) ) {
      Add_SysTime_Str( cp, &st );
   } else {
      strcat(cp, "??/??/???? ??:??:??.????");
   }
   strcat(cp, " "); // add space
   ul.HighPart = pf->nFileSizeHigh;
   ul.LowPart  = pf->nFileSizeLow;
   PTSTR ps = GetI64Stg( &ul );
   size_t len = strlen(ps);
   while( len < MMINSIZE ) {
      strcat(cp, " ");
      len++;
   }
   strcat(cp, ps);
   return cp;
}

PTSTR Get_PFILES_Stg( PFILES pf )
{
   static char _s_pfiles_buf[288*2];
   PTSTR cp = _s_pfiles_buf;
   strcpy(cp, Get_FD_Stg( &pf->fd ) );
   strcat(cp, " ");
   strcat(cp, pf->path.c_str());
   return cp;
}

// hh:mm[:sec]
int   get_TIME( char * cpin )
{
   char * cp = strdupe( cpin );
   char * p1 = strchr( cp, ':' );
   int   hr, min, sec;
   sec = 0;
   min = 0;
   if( p1 ) {
      *p1 = 0;
      p1++;
      sec = 0;
      char * p2 = strchr( p1, ':' );
      if( p2 ) {
         *p2 = 0;
         p2++;
         sec = atoi(p2);
      }
      min = atoi(p1);
   }
   hr = atoi(cp);
   bGotTime = 0;
   if(( hr >= 0 ) && ( hr < 24 ) && ( min >= 0 ) && ( min < 60 ) &&
      ( sec >= 0 ) && ( sec < 60 )) {
      Target.wSecond = sec;
      Target.wHour   = hr;
      Target.wMinute = min;
      bGotTime = 1;
   }
   delete cp;
   return bGotTime;
}

int   get_DATE( char * cpin )
{
   char * cp = strdupe( cpin );
   int   sep = ':';
   char * p1, * p2;
   int   day, mth, yr;

   day = mth = yr = 0;
   bGotDate = 0;
   p1 = strchr( cp, sep );
   if( !p1 ) {
      sep = '/';
      p1 = strchr(cp, sep);
   }
   if(p1) {
      *p1 = 0;
      p1++;
      p2 = strchr(p1, sep);
      if(p2) {
         *p2 = 0;
         p2++;
         day = atoi(cp);
         mth = atoi(p1);
         yr  = atoi(p2);
         if(( day >= 1 ) && ( day < 32 ) && ( mth >= 1 ) && ( mth <= 12 ) &&
            ( yr >= 1 ) ) {
            bGotDate = 1;
            Target.wDay = day;
            Target.wMonth = mth;
            Target.wYear = yr;
         }
      }
   }

   delete cp;
   return bGotDate;
}

int   Get_REF( char * file )
{
   WIN32_FIND_DATA   fd;
   SYSTEMTIME  st;
   HANDLE hFind = FindFirstFile( file, &fd );
   if( VFH(hFind) ) {
      FindClose(hFind);
      if( FileTimeToSystemTime( &fd.ftLastWriteTime, &st ) ) {
         memcpy( &Target, &st, sizeof(SYSTEMTIME) );
         bGotRef = 1;
         return 1;
      }
   }
   return 0;
}

/* ===============================
   code not used

bool Is_Wild( const char * cp )
{
   size_t len = strlen(cp);
   size_t i;
   for( i = 0; i < len; i++ )
   {
      int c = cp[i];
      if(( c == '?' )||( c == '*' ))
         return true;

   }
   return false;
}

void Add_Wild_Files( void )
{
   WIN32_FIND_DATA   fd;
   HANDLE hfind = FindFirstFile( fullpath.path, &fd );
   _splitpath( fullpath.path, fullpath.drive, fullpath.dir, fullpath.fname, fullpath.ext );
   if( hfind && (hfind != INVALID_HANDLE_VALUE) )
   {
      do
      {
         if( !( fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) )
         {
            strcpy( fullpath.path, fullpath.drive );
            strcat( fullpath.path, fullpath.dir );
            strcat( fullpath.path, fd.cFileName );
            string s = fullpath.path;
            files.push_back(s);
         }
      } while( FindNextFile( hfind, &fd ) );
      CloseHandle(hfind);

   }
}
   ================================= */

void get_files(char * cp)
{
   string s = cp;
   if( _fullpath( fullpath.path, cp, _MAX_PATH ) ) {
      s = fullpath.path;
   } else {
      strcpy( fullpath.path, cp );
   }
   //if( Is_Wild( s.c_str() ) )
   //{
   //   Add_Wild_Files();
   //}
   //else
   //{
      files.push_back(s);  // add to string set
   //}
}

void  Process_Input( int argc, char * argv[], string & cmd )
{
   int   i;
   for( i = 1; i < argc; i++ )
   {
      char * cp = argv[i];
      int   c = *cp;
      if( cmd.length() )
         cmd += " ";
      cmd += cp;
      if(( c== '-' ) || ( c == '/' )) {
         cp++;
         c = toupper(*cp);
         switch(c)
         {
         case '?':
            give_help();
            break;
         case 'S':
            bUseSystem = 1;
            break;
         case 'T':
            // get TIME (* = leave current)
            cp++;
            if( !get_TIME( cp ) ) {
               printf( "ERROR: Invalid TIME %s! ...\n", argv[i] );
               pgm_exit(1);
            }
            break;
         case 'U':
            bDoUpdate = 1;
            break;
         case 'D':
            // get DATE (* = leave as current)
            cp++;
            if( !get_DATE( cp ) ) {
               printf( "ERROR: Invalid DATE %s! ...\n", argv[i] );
               pgm_exit(1);
            }
            break;
         case 'R':
            // reference file - us its time/date for update
            cp++;
            if( Reference ) {
               printf( "ERROR: Already have REFERENCE FILE %s ...\n", Reference );
               pgm_exit(1);
            }
            Reference = strdupe(cp);
            if( !Get_REF( cp ) ) {

            }
            break;
         default:
            err_cmd( argv[i] );
            break;
         }

      } else {
         get_files(cp);
      }
   }
}

int   Got_End_Path_Sep( PTSTR path )
{
   size_t len = strlen(path);
   if(len) {
      len--;   // back to last char
      if( strchr( "\\/", path[len] ) )
         return 1;
   }
   return 0;
}

int   Got_Wild( PTSTR file )
{
   size_t len = strlen(file);
   size_t i;
   for(i = 0; i < len; i++)
   {
      int   c = file[i];
      if(( c == '?' )||( c == '*' ))
         return 1;
   }
   return 0;
}

void Add_2_List( PTSTR path, WIN32_FIND_DATA * pfd )
{
   PFILES   pf = new FILES;
   CHKMEM(pf);
   ZeroMemory( &fullpath, sizeof(fullpath) );
   _splitpath( path, fullpath.drive, fullpath.dir, fullpath.fname, fullpath.ext );
   pf->path = fullpath.drive;
   pf->path += fullpath.dir;
   memcpy( &pf->fd, pfd, sizeof(WIN32_FIND_DATA));
   sfiles.push_back(pf);
}

void Process_Directory( PTSTR path )
{
   string s = path;
   int i = Got_End_Path_Sep(path);
   if(!i) s += "\\";
   s += "*.*";
   WIN32_FIND_DATA   fd;
   HANDLE hFind = FindFirstFile( s.c_str(), &fd ); 
   if( VFH(hFind) ) {
      do
      {
         if( !(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ) {
            PFILES   pf = new FILES;
            pf->path = path;
            if(!i) pf->path += "\\";
            memcpy( &pf->fd, &fd, sizeof(WIN32_FIND_DATA));
            sfiles.push_back(pf);
         }
      } while( FindNextFile( hFind, &fd ) );
      FindClose( hFind );
   }
}

int   Check_Do_Update( PTSTR msg )
{
   printf( msg );
   // int chr = cin.get();
   int chr = toupper(_getch());
   fflush(stdin);
   return chr;
}

void  Process_File( string & s )
{
   //string s = *it;
   WIN32_FIND_DATA   fd;
   PTSTR ps = (PTSTR)s.c_str();
   HANDLE hFind = FindFirstFile( ps, &fd ); 
   if( !VFH(hFind) ) {
      printf( "ERROR: %s is NOT a valid file, or directory ...\n", ps );
      pgm_exit(1);
   }
   if( fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
      FindClose( hFind );
      Process_Directory( ps );
   } else if( Got_Wild( ps ) ) {
      do
      {
         if( fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
            // do nothing
         } else {
            Add_2_List( ps, &fd );
         }
      } while( FindNextFile( hFind, &fd ) );
      FindClose( hFind );
   } else {
      FindClose( hFind );
      Add_2_List( ps, &fd );
   }
}

string Get_CWD_Stg( void )
{
   TCHAR _s_CWD_Buf[MAX_PATH];
   PTSTR ps = _s_CWD_Buf;
   _getcwd( ps, MAX_PATH );
   string s = ps;
   return s;
}

void Do_Update( void )
{
   DWORD dwErr;
   int cnt = 0;
   string s;
   for( fiter fit = sfiles.begin(); fit != sfiles.end(); fit++ ) {
      PFILES pf = *fit;
      s = pf->path;
      s += pf->fd.cFileName;
      HANDLE hf = CreateFile( s.c_str(),
         GENERIC_WRITE, 0, NULL, OPEN_EXISTING,
         FILE_ATTRIBUTE_NORMAL, NULL);
      cnt++;
      if( VFH(hf) ) {
         if( SetFileTime( hf, &FileTime, &FileTime, &FileTime ) ) {
            printf( "%d Done SetFileTime on file %s ...\n",
               cnt, pf->fd.cFileName );
         } else {
            dwErr = GetLastError();
            ErrMessage[0] = 0;
            printf( "%d WARNING: SetFileTime on file %s FAILED!\n",
               cnt, pf->fd.cFileName );
            if( GetLastErrorMsg( ErrMessage, 1024, dwErr ) != -1 ) {
               printf( "ERROR Indication: %s ...\n", ErrMessage ); 
            }
         }
         CloseHandle(hf);      
      } else {
         dwErr = GetLastError();
         ErrMessage[0] = 0;
         printf( "%d WARNING: Unable to OPEN file %s to change date/time!\n",
            cnt, pf->fd.cFileName );
         if( GetLastErrorMsg( ErrMessage, 1024, dwErr ) != -1 ) {
            printf( "ERROR Indication: %s ...\n", ErrMessage ); 
         }
      }
   }
}

int main(int argc, char * argv[])
{
   SYSTEMTIME  syst, loct;
   string s = "";

   Process_Input( argc, argv, s );
   printf( "Processed input: %s\n", s.c_str() );

   if( !files.size() ) {
      //no_input();
      files.push_back( Get_CWD_Stg() );
   }
   GetLocalTime( &loct );
   printf( "Current local  time is %s ...\n", Get_SysTime_Stg( &loct ) );
   GetSystemTime( &syst );
   printf( "Current system time is %s ...\n", Get_SysTime_Stg( &syst ) );
   if( bGotRef ) {
      printf( "Using Reference file %s for date and time ...\n", Reference );
   } else if( !bGotTime && !bGotDate ) {
      printf( "No Date, Time or Reference input - " );
      if( bUseSystem )
      {
         memcpy( &Target, &syst, sizeof(SYSTEMTIME) );
         printf( "using system date/time ...\n" );
      }
      else
      {
         memcpy( &Target, &loct, sizeof(SYSTEMTIME) );
         printf( "using local date/time ...\n" );
      }
   } else {
      if( !bGotTime ) {
         if( bUseSystem ) {
            printf( "Using system time, on user date ...\n" );
            Target.wHour = syst.wHour;
            Target.wMinute = syst.wMinute;
            Target.wSecond = syst.wSecond;
            Target.wMilliseconds = syst.wMilliseconds;
            Target.wDayOfWeek = syst.wDayOfWeek;
         } else {
            printf( "Using local time, on user date ...\n" );
            Target.wHour = loct.wHour;
            Target.wMinute = loct.wMinute;
            Target.wSecond = loct.wSecond;
            Target.wMilliseconds = loct.wMilliseconds;
            Target.wDayOfWeek = loct.wDayOfWeek;
         }
      } else if( !bGotDate ) {
         if( bUseSystem ) {
            printf( "Using system date, on users time ...\n" );
            Target.wDay = syst.wDay;
            Target.wMonth = syst.wMonth;
            Target.wYear = syst.wYear;
         } else {
            printf( "Using local date, on users time ...\n" );
            Target.wDay = loct.wDay;
            Target.wMonth = loct.wMonth;
            Target.wYear = loct.wYear;
         }
      } else {
         printf( "Using user input date and time ...\n" );
      }
   }

   printf( "On update, will use %s ...\n", Get_SysTime_Stg( &Target ) );

   if( !SystemTimeToFileTime( &Target, &FileTime ) ) {
      printf( "ERROR: FAILED to convert system time to file time ...\n" );
      pgm_exit(1);
   }

   for( siter it = files.begin(); it != files.end(); it++ ) {
      Process_File( *it );
   }
   size_t cnt = sfiles.size();
   printf( "Found %d files ...\n", cnt );
   cnt = 0;
   for( fiter fit = sfiles.begin(); fit != sfiles.end(); fit++ )
   {
      PFILES pf = *fit;
      cnt++;
      printf( "%s (%d)\n", Get_PFILES_Stg(pf), cnt);
   }

   if( bDoUpdate ) {
      int   up = Check_Do_Update( "Continue with update? Y/N : " );
      if( toupper(up) == 'Y' ) {
         printf( "Yes.\n" );
         Do_Update();
      } else {
         printf( "NO.\n" );
      }
   } else {
      printf( "WARNING: No update done, since no -u command located ...\n" );
   }

   pgm_exit(0);
	return 0;

}

